--------------------------------------------------------------------------------
-- copyright notice:
--------------------------------------------------------------------------------
-- copyright (c) 2007 by
--
-- CESYS GmbH / GERMANY
--
-- http://www.cesys.com
--------------------------------------------------------------------------------
-- author:
--------------------------------------------------------------------------------
-- Stefan Fischer <stefan.fischer@cesys.com>
--------------------------------------------------------------------------------
-- license:
--------------------------------------------------------------------------------
-- THIS SOURCECODE IS NOT FREE! IT IS FOR USE TOGETHER WITH THE CESYS EFM01
-- USB CARD (ARTICLE-NR.: C1050-4107) ONLY! YOU ARE NOT ALLOWED TO MODIFY AND
-- DISTRIBUTE OR USE IT WITH ANY OTHER HARDWARE, SOFTWARE OR ANY OTHER KIND OF
-- ASIC OR PROGRAMMABLE LOGIC DESIGN WITHOUT THE EXPLICIT PERMISSION OF THE
-- COPYRIGHT HOLDER!
--------------------------------------------------------------------------------
-- disclaimer of warranty:
--------------------------------------------------------------------------------
-- THIS SOURCECODE IS DISTRIBUTED IN THE HOPE THAT IT WILL BE USEFUL, BUT
-- THERE IS NO WARRANTY OR SUPPORT FOR THIS SOURCECODE. THE COPYRIGHT HOLDER 
-- PROVIDES THIS SOURCECODE "AS IS" WITHOUT WARRANTY OF ANY KIND,
-- EITHER EXPRESSED OR IMPLIED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED 
-- WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
-- THE ENTIRE RISK AS TO THE QUALITY AND PERFORMANCE OF THIS SOURCECODE IS WITH
-- YOU. SHOULD THIS SOURCECODE PROVE DEFECTIVE, YOU ASSUME THE COST OF ALL 
-- NECESSARY SERVICING, REPAIR OR CORRECTION.
--
-- IN NO EVENT WILL THE COPYRIGHT HOLDER BE LIABLE TO YOU FOR DAMAGES, INCLUDING
-- ANY GENERAL, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES ARISING OUT OF THE
-- USE OR INABILITY TO USE THIS SOURCECODE (INCLUDING BUT NOT LIMITED TO LOSS OF
-- DATA OR DATA BEING RENDERED INACCURATE OR LOSSES SUSTAINED BY YOU OR THIRD 
-- PARTIES OR A FAILURE OF THIS SOURCECODE TO OPERATE WITH ANY OTHER
-- SOFTWARE-PROGRAMS, HARDWARE-CIRCUITS OR ANY OTHER KIND OF ASIC OR
-- PROGRAMMABLE LOGIC DESIGN), EVEN IF THE COPYRIGHT HOLDER HAS BEEN ADVISED OF 
-- THE POSSIBILITY OF SUCH DAMAGES.
--------------------------------------------------------------------------------


library IEEE;
use IEEE.STD_LOGIC_1164.all;
use IEEE.NUMERIC_STD.all;

-- pragma translate_off
library UNISIM;
use UNISIM.VCOMPONENTS.all;
-- pragma translate_on


entity sync_fifo16 is
  generic
  (
    data_width : positive := 16;
    use_xil_dram : boolean := false;
    nr_of_dbgports : positive := 1
  );
  port
  (
    rst : in std_logic;
    clk : in std_logic;

    wr_i : in std_logic;
    wr_data_i : in std_logic_vector((data_width-1) downto 0);
    wr_full_o : out std_logic;

    rd_i : in std_logic;
    rd_data_o : out std_logic_vector((data_width-1) downto 0);
    rd_empty_o : out std_logic;

    count_o : out std_logic_vector(3 downto 0);

    debug : inout std_logic_vector((nr_of_dbgports-1) downto 0)
  );
end sync_fifo16;


architecture RTL of sync_fifo16 is

  signal wr : std_logic := '0';
  signal wr_data : std_logic_vector((data_width-1) downto 0) := (others => '0');
  signal wr_full : std_logic := '1';

  signal rd : std_logic := '0';
  signal rd_data : std_logic_vector((data_width-1) downto 0) := (others => '0');
  signal rd_empty : std_logic := '1';

  signal count : std_logic_vector(3 downto 0) := (others => '0');

  ----- component RAM16X1D -----
  component RAM16X1D
    generic
    (
      INIT : bit_vector(15 downto 0) := X"0000"
    );
    port
    (
      DPO : out std_ulogic;
      SPO : out std_ulogic;
      A0 : in std_ulogic;
      A1 : in std_ulogic;
      A2 : in std_ulogic;
      A3 : in std_ulogic;
      D : in std_ulogic;
      DPRA0 : in std_ulogic;
      DPRA1 : in std_ulogic;
      DPRA2 : in std_ulogic;
      DPRA3 : in std_ulogic;
      WCLK : in std_ulogic;
      WE : in std_ulogic
    );
  end component;
    
	signal DATA_IN : std_logic_vector((data_width-1) downto 0) := (others => '0');
  signal ADDRESS : std_logic_vector(3 downto 0) := (others => '0');
  signal ADDRESS_DP : std_logic_vector(3 downto 0) := (others => '0');
  signal WRITE_EN : std_logic := '0';
  signal READ_EN : std_logic := '0';
  signal DATA_OUT : std_logic_vector((data_width-1) downto 0) := (others => '0');
  signal DATA_OUT_DP : std_logic_vector((data_width-1) downto 0) := (others => '0');
  signal O_DATA_OUT : std_logic_vector((data_width-1) downto 0) := (others => '0');
  signal O_DATA_OUT_DP : std_logic_vector((data_width-1) downto 0) := (others => '0');
  
  type arr_std is array(natural range <>) of std_logic_vector((data_width-1) downto 0);
  signal fifo_buff : arr_std(15 downto 0) := (others =>(others => '0'));
  signal fifo_buff_regout : std_logic_vector((data_width-1) downto 0) := (others => '0');

  signal wr_en : std_logic := '0';
  signal wr_pointer : unsigned(3 downto 0) := (others => '0');
  signal rd_en : std_logic := '0';
  signal rd_pointer : unsigned(3 downto 0) := (others => '0');

begin

  wr <= wr_i;
  wr_data <= wr_data_i;
  wr_full_o <= wr_full;
  
  rd <= rd_i;
  rd_data_o <= rd_data;
  rd_empty_o <= rd_empty;

  count_o <= count;
  
  debug <=
  (
    others => 'Z'
  );

  inst_xil_dram:
  if use_xil_dram generate
  
    RAM16X1D_X:
    for i in 0 to (data_width-1) generate
      -- RAM16X1D: 16 x 1 positive edge write, asynchronous read dual-port distributed RAM
      --           All FPGAs
      -- Xilinx HDL Language Template, version 9.2.1i
    
      RAM16X1D_inst : RAM16X1D
      generic map (
         INIT => X"0000")
      port map (
         DPO => DATA_OUT_DP(i),     -- Read-only 1-bit data output for DPRA
         SPO => DATA_OUT(i),     -- R/W 1-bit data output for A0-A3
         A0 => ADDRESS(0),       -- R/W address[0] input bit
         A1 => ADDRESS(1),       -- R/W address[1] input bit
         A2 => ADDRESS(2),       -- R/W address[2] input bit
         A3 => ADDRESS(3),       -- R/W ddress[3] input bit
         D => DATA_IN(i),         -- Write 1-bit data input
         DPRA0 => ADDRESS_DP(0), -- Read-only address[0] input bit
         DPRA1 => ADDRESS_DP(1), -- Read-only address[1] input bit
         DPRA2 => ADDRESS_DP(2), -- Read-only address[2] input bit
         DPRA3 => ADDRESS_DP(3), -- Read-only address[3] input bit
         WCLK => clk,   -- Write clock input
         WE => WRITE_EN        -- Write enable input
      );
    
      -- End of RAM16X1D_inst instantiation
    end generate;
    
  end generate;

	DATA_IN <= wr_data;
  ADDRESS <= std_logic_vector(wr_pointer);
  ADDRESS_DP <= std_logic_vector(rd_pointer);
  WRITE_EN <= wr_en;
  READ_EN <= rd_en;
  rd_data	<= O_DATA_OUT_DP when use_xil_dram else fifo_buff_regout;

  wr_en <= wr and (not wr_full);
  rd_en <= rd and (not rd_empty);

  process(rst, clk)
    variable wr_rd : std_logic_vector(1 downto 0) := (others => '0');
  begin
    if rst = '1' then
      wr_pointer <= (others => '0');
      wr_full <= '1';
      rd_pointer <= (others => '0');
      rd_empty <= '1';
      count <= (others => '0');
      if use_xil_dram then
        O_DATA_OUT <= (others => '0');
        O_DATA_OUT_DP <= (others => '0');
      else
        fifo_buff <= (others => (others => '0'));
        fifo_buff_regout <= (others => '0');
      end if;
    elsif clk'event and clk = '1' then
      if wr_en = '1' then
        wr_pointer <= wr_pointer + 1;
      end if;
      if rd_en = '1' then
        rd_pointer <= rd_pointer + 1;
      end if;
      wr_rd := wr_en & rd_en;
      -- empty-flag
      if (unsigned(count) = 0) or (unsigned(count) = 1 and wr_rd = b"01") then
        rd_empty <= '1';
      else
        rd_empty <= '0';
      end if;
      -- full-flag
      if (unsigned(count) = 15) or (unsigned(count) = 14 and wr_rd = b"10") then
        wr_full <= '1';
      else
        wr_full <= '0';
      end if;
      -- count
      case wr_rd is
        when b"01" =>
          count <= std_logic_vector(unsigned(count) - 1);
        when b"10" =>
          count <= std_logic_vector(unsigned(count) + 1);
        when others => null;
      end case;
      if use_xil_dram then
        if READ_EN = '1' then
          O_DATA_OUT <= DATA_OUT;
          O_DATA_OUT_DP <= DATA_OUT_DP;
        end if;
      else
        if wr_en = '1' then
          fifo_buff(TO_INTEGER(wr_pointer)) <= wr_data;
        end if;
        if rd_en = '1' then
          fifo_buff_regout <= fifo_buff(TO_INTEGER(rd_pointer));
        end if;
      end if;
    end if;
  end process;

end RTL;